/*
 * Copyright (C) 2021 Purism SPC
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * Author: Evangelos Ribeiro Tzaras <devrtz@fortysixandtwo.eu>
 *
 */

#pragma once

#include "calls-contacts-provider.h"

CallsContactsProvider *__wrap_calls_contacts_provider_new (CallsSettings *settings);
